/* 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.muse.ws.addressing.soap;

import javax.xml.namespace.QName;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.muse.util.xml.XmlSerializable;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.addressing.soap.SoapConstants;

/**
 * 
 * This class represents a the SOAP fault type defined in SOAP 1.2.
 * 
 * edited by Matthes Rieke to fulfill ows service needs.
 * 
 * @author Dan Jemiolo (danj)
 *
 */

public class SoapFault extends Exception implements XmlSerializable
{
	private static final long serialVersionUID = -8612773172898054578L;

	private QName _code = SoapConstants.RECEIVER_QNAME;

	private QName _subCode = null;
	private String _reason = null;
	private String _nodeURI = null;
	private String _roleURI = null;

	private Element _detail = null;

	public SoapFault(Element xml)
	{
		Element codeXML = XmlUtils.getElement(xml, SoapConstants.CODE_QNAME);
		_code = XmlUtils.getQNameFromChild(codeXML, SoapConstants.VALUE_QNAME);

		Element subCodeXML = XmlUtils.getElement(codeXML, SoapConstants.SUBCODE_QNAME);

		if (subCodeXML != null)
			_subCode = XmlUtils.getQNameFromChild(subCodeXML, SoapConstants.VALUE_QNAME);

		Element reasonXML = XmlUtils.getElement(xml, SoapConstants.REASON_QNAME);
		_reason = XmlUtils.getElementText(reasonXML, SoapConstants.TEXT_QNAME);

		_nodeURI = XmlUtils.getElementText(xml, SoapConstants.NODE_QNAME);
		_roleURI = XmlUtils.getElementText(xml, SoapConstants.ROLE_QNAME);

		Element detailXML = XmlUtils.getElement(xml, SoapConstants.DETAIL_QNAME);

		if (detailXML != null)
			_detail = XmlUtils.getFirstElement(detailXML);
	}

	public SoapFault(String message)
	{
		super(message);
		_reason = message;
	}

	public SoapFault(String message, Throwable cause)
	{
		super(message, cause);
		_reason = message;     
	}

	public SoapFault(Throwable cause)
	{
		super(cause);
		_reason = cause.getMessage();
	}

	public QName getCode()
	{
		return _code;
	}

	public void setCode(QName code)
	{
		_code = code;
	}

	public Element getDetail()
	{
		return _detail;
	}

	public void setDetail(Element detail)
	{
		_detail = detail;
	}

	public String getMessage()
	{
		return getReason();
	}

	public String getNodeURI()
	{
		return _nodeURI;
	}

	public void setNodeURI(String nodeURI)
	{
		_nodeURI = nodeURI;
	}

	public String getReason()
	{
		return _reason;
	}

	public void setReason(String reason)
	{
		_reason = reason;
	}

	public String getRoleURI()
	{
		return _roleURI;
	}

	public void setRoleURI(String roleURI)
	{
		_roleURI = roleURI;
	}

	public QName getSubCode()
	{
		return _subCode;
	}

	public void setSubCode(QName subCode)
	{
		_subCode = subCode;
	}

	public Element toXML()
	{
		return toXML(XmlUtils.createDocument());
	}

	public Element toXML(Document doc)
	{
		Element faultXML = XmlUtils.createElement(doc, SoapConstants.FAULT_QNAME);

		Element codeXML = XmlUtils.createElement(doc, SoapConstants.CODE_QNAME);
		faultXML.appendChild(codeXML);

		XmlUtils.setElement(codeXML, SoapConstants.VALUE_QNAME, getCode());

		QName subCode = getSubCode();

		Element detail = getDetail();

		if (detail != null) {
			setReason(getReason()+ XmlUtils.toString(detail, false));
		}

		if (subCode != null)
		{
			Element subCodeXML = XmlUtils.createElement(doc, SoapConstants.SUBCODE_QNAME);
			codeXML.appendChild(subCodeXML);
			XmlUtils.setElement(subCodeXML, SoapConstants.VALUE_QNAME, subCode);
		}

		Element reasonXML = XmlUtils.createElement(doc, SoapConstants.REASON_QNAME);
		Element reasonTextXML = XmlUtils.createElement(doc, SoapConstants.TEXT_QNAME, getReason());
		reasonTextXML.setAttributeNS("http://www.w3.org/XML/1998/namespace", "xml:lang", "en-US");
		reasonXML.appendChild(reasonTextXML);
		faultXML.appendChild(reasonXML);

		String nodeURI = getNodeURI();

		if (nodeURI != null)
			XmlUtils.setElement(faultXML, SoapConstants.NODE_QNAME, nodeURI); 

		String roleURI = getRoleURI();

		if (roleURI != null)
			XmlUtils.setElement(faultXML, SoapConstants.ROLE_QNAME, roleURI); 

		Element excRep = XmlUtils.createElement(new QName("http://www.opengis.net/ows", "ExceptionReport"));
		excRep.setAttribute("version", "0.0.0");
		Element exc = XmlUtils.createElement(new QName("http://www.opengis.net/ows", "Exception"));
		exc.setAttribute("exceptionCode", "Unspecified");
		Element excText = XmlUtils.createElement(new QName("http://www.opengis.net/ows", "ExceptionText"),
				doc.createCDATASection(convertStacktractToString(getCause())));
		exc.appendChild(excText);
		excRep.appendChild(exc);

		setDetail(excRep);

		XmlUtils.setElement(faultXML, SoapConstants.DETAIL_QNAME, excRep, false);

		return faultXML;
	}

	private String convertStacktractToString(Throwable throwable) {
		if (throwable == null) return "Exception was null.";
		
		StringBuilder sb = new StringBuilder();
		sb.append(throwable.getMessage());
		sb.append(System.getProperty("line.separator"));
		for (StackTraceElement stackTraceElement : throwable.getStackTrace()) {
			sb.append(stackTraceElement);
			sb.append(System.getProperty("line.separator"));
		}
		return sb.toString();
	}
}
